/* $Header: /var/cvs/mbdyn/mbdyn/mbdyn-1.0/libraries/libmbutil/parser.cc,v 1.53 2009/07/13 10:42:54 masarati Exp $ */
/*
 * MBDyn (C) is a multibody analysis code.
 * http://www.mbdyn.org
 *
 * Copyright (C) 1996-2009
 *
 * Pierangelo Masarati	<masarati@aero.polimi.it>
 * Paolo Mantegazza	<mantegazza@aero.polimi.it>
 *
 * Dipartimento di Ingegneria Aerospaziale - Politecnico di Milano
 * via La Masa, 34 - 20156 Milano, Italy
 * http://www.aero.polimi.it
 *
 * Changing this copyright notice is forbidden.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation (version 2 of the License).
 *
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* parser */

#ifdef HAVE_CONFIG_H
#include <mbconfig.h>           /* This goes first in every *.c,*.cc file */
#endif /* HAVE_CONFIG_H */

#include <stdlib.h>

#include "parser.h"
#include "Rot.hh"

/* LowParser - begin */

static int
skip_remarks(HighParser& HP, InputStream& In, char &cIn)
{
skip_again:;
	for (cIn = In.get(); isspace(cIn); cIn = In.get()) {
		if (In.eof()) {
			return -1;
		}
	}

	switch (cIn) {
	case MathParser::ONE_LINE_REMARK:
		for (cIn = In.get(); cIn != '\n'; cIn = In.get()) {
			if (cIn == '\\') {
				cIn = In.get();
				if (In.eof()) {
					return -1;
				}
				if (cIn == '\r') {
					/* if input file was prepared
					 * under DOS/Windows */
					cIn = In.get();
					if (In.eof()) {
						return -1;
					}
				}
			}
			if (In.eof()) {
				return -1;
			}
		}
		goto skip_again;

	case '/':
		cIn = In.get();
		if (In.eof()) {
			return -1;

		} else if (cIn == '*') {
			for (cIn = In.get(); !In.eof(); cIn = In.get()) {
				if (cIn == '*') {
end_of_comment:;
					cIn = In.get();
					if (In.eof()) {
						return -1;
					}
					if (cIn == '/') {
						goto skip_again;
					}

				} else if (cIn == '/') {
					cIn = In.get();
					if (In.eof()) {
						return -1;
					}
					if (cIn == '*') {
						silent_cerr("warning: '/*' inside a comment "
							"at line " << HP.GetLineData()
							<< std::endl);
						goto end_of_comment;
					}
				}
			}
			if (In.eof()) {
				return -1;
			}

		} else {
			In.putback(cIn);
			return 0;
		}
	}

	return 0;
}

LowParser::LowParser(HighParser& hp)
: HP(hp), sCurrWordBuf(0), iBufSize(iDefaultBufSize)
{
	SAFENEWARR(sCurrWordBuf, char, iBufSize);
}

LowParser::~LowParser(void)
{
	if (sCurrWordBuf) {
		SAFEDELETEARR(sCurrWordBuf);
	}
}

void
LowParser::PackWords(InputStream& In)
{
	unsigned iCur = 0;
	char cIn;

	/* note: no remarks allowed inside words */
	for (cIn = In.get(); !In.eof(); cIn = In.get()) {
		switch (cIn) {
		case COLON:
		case COMMA:
		case SEMICOLON:
			goto end_of_word;

		default:
      			if (!isspace(cIn)) {
				sCurrWordBuf[iCur] = cIn;
				iCur++;
				 if (iCur == iBufSize - 1) {
					 char *s = NULL;
					 unsigned i = 2*iBufSize;

					 /* FIXME: no limit on max size? */

					 SAFENEWARR(s, char, i);
					 memcpy(s, sCurrWordBuf, iBufSize);
					 SAFEDELETEARR(sCurrWordBuf);
					 sCurrWordBuf = s;
					 iBufSize = i;
				 }
			}
		}
	}

	throw EndOfFile(MBDYN_EXCEPT_ARGS);

end_of_word:;

	sCurrWordBuf[iCur] = '\0';
	In.putback(cIn);
}


LowParser::Token
LowParser::GetToken(InputStream& In)
{
	/* toglie gli spazi iniziali e tutti i commenti */
	char cIn;
	if (skip_remarks(HP, In, cIn)) {
		return CurrToken = LowParser::ENDOFFILE;
	}

	if (isalpha(cIn) || cIn == '_') {
		PackWords(In.putback(cIn));
		return CurrToken = LowParser::WORD;
	}

	switch (cIn) {
	case ',':
		return CurrToken = LowParser::COMMA;

	case ':':
		return CurrToken = LowParser::COLON;

	case ';':
		return CurrToken = LowParser::SEMICOLON;

	case '.':
	case '-':
	case '+':
is_digit:;
		In.putback(cIn) >> dCurrNumber;
		return CurrToken = LowParser::NUMBER;

	default:
		if (isdigit(cIn)) {
			goto is_digit;
		}
		In.putback(cIn);
		return CurrToken = LowParser::UNKNOWN;
	}
}


doublereal
LowParser::dGetReal(void)
{
	return dCurrNumber;
}


integer
LowParser::iGetInt(void)
{
	return integer(dCurrNumber);
}


char*
LowParser::sGetWord(void)
{
	return sCurrWordBuf;
}

/* LowParser - end */


/* KeyTable - begin */

KeyTable::KeyTable(HighParser& hp, const char* const sTable[])
: sKeyWords(0), oldKey(0), HP(hp) 
{
	sKeyWords = (char* const*)sTable;
	oldKey = HP.PutKeyTable(*this);
}


KeyTable::~KeyTable(void)
{
	if (oldKey) {
		(void)HP.PutKeyTable(*oldKey);
	}
}


int
KeyTable::Find(const char* sToFind) const
{
	for (int iCnt = 0; sKeyWords[iCnt]; iCnt++) {
		if (strcasecmp(sKeyWords[iCnt], sToFind) == 0) {
			return iCnt;
		}
	}

	return -1;
}

/* KeyTable - end */


/* HighParser - begin */

HighParser::HighParser(MathParser& MP, InputStream& streamIn)
: ESCAPE_CHAR('\\'),
LowP(*this),
pIn(&streamIn),
pf(NULL),
MathP(MP),
KeyT(0)
{
	DEBUGCOUTFNAME("HighParser::HighParser");
	CurrToken = HighParser::DESCRIPTION;
}


HighParser::~HighParser(void)
{
	DEBUGCOUTFNAME("HighParser::~HighParser");
	Close();
}


void
HighParser::Close(void)
{
	NO_OP;
}


const KeyTable*
HighParser::PutKeyTable(const KeyTable& KT)
{
	const KeyTable* oldKey = KeyT;

	KeyT = &KT;

	return oldKey;
}

MathParser&
HighParser::GetMathParser(void)
{
	return MathP;
}

int
HighParser::GetLineNumber(void) const
{
	return ((InputStream*)pIn)->GetLineNumber();
}


HighParser::ErrOut
HighParser::GetLineData(void) const
{
	ErrOut LineData;
	LineData.iLineNumber = GetLineNumber();
	LineData.sFileName = NULL;
	return LineData;
}


bool
HighParser::IsDescription(void)
{
     	if (CurrToken != HighParser::DESCRIPTION) {
	  	return false;
     	}
     	return true;
}


int
HighParser::iGetDescription_int(const char* const s)
{
	int i = -1;
	
	if (KeyT) {
		i = KeyT->Find(s);
	}

     	if (FirstToken() == HighParser::UNKNOWN) {
		silent_cerr("Parser error in HighParser::iGetDescription_int(), "
			"semicolon expected at line "
			<< GetLineData() << std::endl);
	  	throw HighParser::ErrSemicolonExpected(MBDYN_EXCEPT_ARGS);
     	}

     	return i;
}


void
HighParser::Set_int(void)
{
     	if (FirstToken() == UNKNOWN) {
     		silent_cerr("Parser error in HighParser::Set_int(), "
     			"colon expected at line "
     			<< GetLineData() << std::endl);
     		throw HighParser::ErrColonExpected(MBDYN_EXCEPT_ARGS);
	}

	TypedValue v;
	GetValue(v);
}

void
HighParser::SetEnv_int(void)
{
#ifdef HAVE_SETENV
     	if (FirstToken() == UNKNOWN) {
     		silent_cerr("Parser error in HighParser::SetEnv_int(), "
     			"colon expected at line "
     			<< GetLineData() << std::endl);
     		throw HighParser::ErrColonExpected(MBDYN_EXCEPT_ARGS);
	}

	int overwrite = 0;
	if (IsKeyWord("overwrite")) {
		overwrite = 1;
	}

	const char *ava = GetStringWithDelims();
	if (ava == NULL) {
		silent_cerr("unable to get AVA for \"setenv\" at line "
				<< GetLineData() << std::endl);
		throw ErrGeneric(MBDYN_EXCEPT_ARGS);
	}

	char *avasep = strchr(ava, '=');
	if (avasep == NULL) {
		int	rc = 0;
#ifdef HAVE_UNSETENV
		unsetenv(ava);
#elif defined(HAVE_PUTENV)
		rc = putenv(ava);
#endif	/* !HAVE_UNSETENV && !HAVE_PUTENV */
		if (rc) {
			silent_cerr("unable to unset the environment variable "
					"\"" << ava << "\" at line "
					<< GetLineData() << std::endl);
			throw ErrGeneric(MBDYN_EXCEPT_ARGS);
		}


	} else {
		if (avasep == ava) {
			silent_cerr("illegal AVA \"" << ava
					<< "\" at line "
					<< GetLineData() << std::endl);
			throw ErrGeneric(MBDYN_EXCEPT_ARGS);
		}

		avasep[0] = '\0';
		avasep++;
		int rc = setenv(ava, avasep, overwrite);
		if (rc) {
			silent_cerr("unable to set the environment variable \""
					<< ava << "\" to \"" << avasep 
					<< "\" at line " << GetLineData()
					<< std::endl);
			throw ErrGeneric(MBDYN_EXCEPT_ARGS);
		}
	}
#endif // HAVE_SETENV
}

void
HighParser::Remark_int(void)
{
	if (FirstToken() == UNKNOWN) {
		silent_cerr("Parser error in MBDynParser::Remark_int(),"
			" colon expected at line "
			<< GetLineData() << std::endl);
		throw HighParser::ErrColonExpected(MBDYN_EXCEPT_ARGS);
	}

	/* eat it anyway ;) */
	const char *s = GetStringWithDelims();
	silent_cout("line " << GetLineData() << ": " << s);
	while (IsArg()) {
		TypedValue v;
		v = GetValue(v);
		silent_cout(", " << v);
	}

	silent_cout(std::endl);
}


void
HighParser::Eof(void)
{
	 throw EndOfFile(MBDYN_EXCEPT_ARGS);
}

int
HighParser::GetDescription(void)
{
	/* Checks if current token is a description */
	if (!IsDescription()) {
		silent_cerr("Parser error in HighParser::GetDescription, "
			"invalid call to GetDescription at line "
			<< GetLineData() << std::endl);
		throw HighParser::ErrInvalidCallToGetDescription(MBDYN_EXCEPT_ARGS);
	}

restart_parsing:;

	CurrLowToken = LowP.GetToken(*pIn);
	if (CurrLowToken != LowParser::WORD) {
		if (pIn->GetStream().eof()) {
			Eof();
			goto restart_parsing;

		}
		
		silent_cerr("Parser error in HighParser::GetDescription, "
			<< "keyword expected at line "
			<< GetLineData() << std::endl);
		throw HighParser::ErrKeyWordExpected(MBDYN_EXCEPT_ARGS);
	}

	/* Description corrente */
	char* s = LowP.sGetWord();

	if (GetDescription_int(s)) {
		goto restart_parsing;
	}

	return iGetDescription_int(s);
}


/*
 * special descriptions, that are eaten by the parser
 */
bool
HighParser::GetDescription_int(const char *s)
{
	/* Se trova un remark, scrive il commento ed eventualmente
	 * quello che segue */
	if (!strcmp(s, "remark")) {
		Remark_int();
		return true;
		
 	/* display the symbol table */
	} else if (strcmp(s, "print" "symbol" "table" ) == 0 ) {
		/* FIXME: move to a ddicated helper? */
		if (FirstToken() == UNKNOWN) {
			silent_cerr("Parser error in MBDynParser::GetDescription_int(),"
				" semicolon expected at line "
				<< GetLineData() << std::endl);
			throw HighParser::ErrSemicolonExpected(MBDYN_EXCEPT_ARGS);
		}

		silent_cout( "math parser symbol table at line "
				<< GetLineData() << ":" << std::endl
				<< MathP.GetSymbolTable() << std::endl );
		return true;
		
	/* calls the MathParser */
	} else if (strcmp(s, "set") == 0) {
		Set_int();
		return true;

	/* sets environment variable */
	} else if (strcmp(s, "setenv") == 0) {
		SetEnv_int();
		return true;

	/* exits with no error */
	} else if (strcmp(s, "exit") == 0) {
		throw NoErr(MBDYN_EXCEPT_ARGS);
	}

	return false;
}


HighParser::Token
HighParser::FirstToken(void)
{
	CurrLowToken = LowP.GetToken(*pIn);

	switch (CurrLowToken) {
	case LowParser::COLON:
		CurrToken = HighParser::ARG;
		break;

	case LowParser::SEMICOLON:
		CurrToken = HighParser::DESCRIPTION;
		break;

	default:
		CurrToken = HighParser::UNKNOWN;
		break;
	}

	return CurrToken;
}

void
HighParser::ExpectDescription(void)
{
	/* forces the next expected token to be a "description"
	 * e.g. a keyword followed by a colon (deprecated) */
	CurrToken = HighParser::DESCRIPTION;
}


void
HighParser::ExpectArg(void)
{
	/* forces the next expected token to be an argument
	 * e.g. a keyword followed by a separator (deprecated) */
	CurrToken = HighParser::ARG;
}


bool
HighParser::IsArg(void)
{
	if (CurrToken == ARG) {
		return true;
	}
	return false;
}

void
HighParser::PutBackSemicolon(void)
{
	if (CurrLowToken == LowParser::SEMICOLON) {
		pIn->putback(';');
	}
}


void
HighParser::NextToken(const char* sFuncName)
{
	CurrLowToken = LowP.GetToken(*pIn);
	switch (CurrLowToken) {
	case LowParser::COMMA:
		CurrToken = HighParser::ARG;
		break;

	case LowParser::SEMICOLON:
		CurrToken = HighParser::DESCRIPTION;
		break;

	default:
		silent_cerr("Parser error in "
			<< sFuncName << ", missing separator at line "
			<< GetLineData() << std::endl);
		throw HighParser::ErrMissingSeparator(MBDYN_EXCEPT_ARGS);
	}
}

int
HighParser::ParseWord(unsigned flags)
{
	char* sBuf = sStringBuf;
	char* sBufWithSpaces = sStringBufWithSpaces;

	char cIn;
	if (skip_remarks(*this, *pIn, cIn)) {
		return CurrToken = HighParser::ENDOFFILE;
	}

	if (!isalpha(cIn) && cIn != '_') {
		pIn->putback(cIn);
		return -1;
	}

	*sBufWithSpaces++ = cIn;

	if (flags & LOWER) {
		*sBuf++ = tolower(cIn);

	} else if (flags & UPPER) {
		*sBuf++ = toupper(cIn);

	} else {
		*sBuf++ = cIn;
	}

	for (cIn = pIn->get(); isalnum(cIn) || cIn == '_' || isspace(cIn); cIn = pIn->get()) {
		*sBufWithSpaces++ = cIn;
		if (sBufWithSpaces >= sStringBufWithSpaces + iDefaultBufSize - 1) {
			break;
		}

		if (isspace(cIn)) {
			continue;
		}
		
		if (flags & LOWER) {
			*sBuf++ = tolower(cIn);

		} else if (flags & UPPER) {
			*sBuf++ = toupper(cIn);

		} else {
			*sBuf++ = cIn;
		}
	}
	pIn->putback(cIn);

	*sBuf = '\0';
	*sBufWithSpaces = '\0';

	return 0;
}

void
HighParser::PutbackWord(void)
{
	char* sBufWithSpaces = sStringBufWithSpaces + strlen(sStringBufWithSpaces);

   
	while (sBufWithSpaces > sStringBufWithSpaces) {
		pIn->putback(*--sBufWithSpaces);
	}
}

bool
HighParser::IsKeyWord(const char* sKeyWord)
{
	const char sFuncName[] = "HighParser::IsKeyWord()";

	if (CurrToken != HighParser::ARG) {
		return false;
	}

	switch (ParseWord()) {
	case 0:
		break;

	case HighParser::ENDOFFILE:
		return true;

	default:
		return false;
	}

	if (!strcasecmp(sStringBuf, sKeyWord)) {
		NextToken(sFuncName);
		return true;
	}

	PutbackWord();

	return false;
}

int
HighParser::IsKeyWord(void)
{
	const char sFuncName[] = "HighParser::IsKeyWord()";

	if (CurrToken != HighParser::ARG) {
		return -1;
	}

	switch (ParseWord()) {
	case 0:
		break;

	case HighParser::ENDOFFILE:
		return HighParser::ENDOFFILE;

	default:
		return -1;
	}

	int iKW = -1;

	if (KeyT) {
		iKW = KeyT->Find(sStringBuf);
	}
   
	if (iKW >= 0) {
		NextToken(sFuncName);
		return iKW;
	}

	PutbackWord();

	return -1;
}

/* 1 se l'argomento successivo e' una parola in un WordSet */
const char *
HighParser::IsWord(const HighParser::WordSet& ws)
{
	const char sFuncName[] = "HighParser::IsWord()";

	if (CurrToken != HighParser::ARG) {
		return 0;
	}

	switch (ParseWord()) {
	case 0:
		break;

	default:
		return 0;
	}

	if (ws.IsWord(std::string(sStringBuf))) {
		NextToken(sFuncName);
		return sStringBuf;
	}

	PutbackWord();

	return false;

}

TypedValue
HighParser::GetValue(const TypedValue& vDefVal)
{
	const char sFuncName[] = "HighParser::GetValue()";

	if (CurrToken != HighParser::ARG) {
		silent_cerr("Parser error in "
			<< sFuncName << ", arg expected at line "
			<< GetLineData() << std::endl);
		throw HighParser::ErrIntegerExpected(MBDYN_EXCEPT_ARGS);
	}

	TypedValue v(vDefVal);

	try {
		v = MathP.Get(*pIn, v);
	}
	catch (MathParser::ErrGeneric e) {
		silent_cerr(sFuncName << ": error return from MathParser at line "
			<< GetLineData() << std::endl);
		throw e;
	}

	NextToken(sFuncName);
	return v;
}

bool
HighParser::GetBool(bool bDefVal)
{
	TypedValue v(bDefVal);
	v = GetValue(v);
	return v.GetBool();
}

integer
HighParser::GetInt(int iDefVal)
{
	TypedValue v(iDefVal);
	v = GetValue(v);
	return v.GetInt();
}

doublereal
HighParser::GetReal(const doublereal& dDefVal)
{
	TypedValue v(dDefVal);
	v = GetValue(v);
	return v.GetReal();
}

std::string
HighParser::GetString(const std::string& sDefVal)
{
	TypedValue v(sDefVal);
	v = GetValue(v);
	return v.GetString();
}


int
HighParser::GetWord(void)
{
	const char sFuncName[] = "HighParser::GetWord()";

	if (CurrToken != HighParser::ARG) {
		silent_cerr("Parser error in "
			<< sFuncName << ", keyword arg expected at line "
			<< GetLineData() << std::endl);
		throw HighParser::ErrKeyWordExpected(MBDYN_EXCEPT_ARGS);
	}

	CurrLowToken = LowP.GetToken(*pIn);
	if (CurrLowToken != LowParser::WORD) {
		silent_cerr("Parser error in "
			<< sFuncName << ", keyword expected at line "
			<< GetLineData() << std::endl);
		throw HighParser::ErrKeyWordExpected(MBDYN_EXCEPT_ARGS);
	}

	int i = -1;
	if (KeyT) {
		i = KeyT->Find(LowP.sGetWord());
	}

	NextToken(sFuncName);

	return i;
}

const char*
HighParser::GetString(unsigned flags)
{
	const char sFuncName[] = "HighParser::GetString()";

	pedantic_cout("use of deprecated method \"GetString\" at line"
		<< GetLineData() << std::endl);

	if (CurrToken != HighParser::ARG) {
		silent_cerr("Parser error in "
			<< sFuncName << ", string arg expected at line "
			<< GetLineData() << std::endl);
		throw HighParser::ErrStringExpected(MBDYN_EXCEPT_ARGS);
	}

	char* s = sStringBuf;
	char* sTmp = s;

	char cIn = '\0';

	while (isspace(cIn = pIn->get())) {
		NO_OP;
	}

	if (pIn->eof()) {
		CurrToken = HighParser::ENDOFFILE;
		return NULL;
	}

	pIn->putback(cIn);
	for (cIn = pIn->get(); cIn != ',' && cIn != ';'; cIn = pIn->get()) {
		/* Attenzione! cosi' la legge tutta,
		 * ma ne tiene solo iBufSize-1 caratteri */
		if (pIn->eof()) {
			CurrToken = HighParser::ENDOFFILE;
			*sTmp = '\0';
			return s;

		} else if (sTmp < s + iDefaultBufSize - 1) {
			if (!(flags & HighParser::EATSPACES) || !isspace(cIn)) {
				if (flags & HighParser::LOWER) {
					cIn = tolower(cIn);

				} else if (flags & HighParser::UPPER) {
					cIn = toupper(cIn);
				}

				*sTmp++ = cIn;
			}
		}
	}

	pIn->putback(cIn);
	*sTmp = '\0';

	NextToken(sFuncName);

	return s;
}

void
HighParser::SetDelims(enum Delims Del, char &cLdelim, char &cRdelim) const
{
	cLdelim = '\0';
	cRdelim = '\0';

	switch (Del) {
	case PLAINBRACKETS:
		cLdelim = '(';
		cRdelim = ')';
		break;

	case SQUAREBRACKETS:
		cLdelim = '[';
		cRdelim = ']';
		break;

	case CURLYBRACKETS:
		cLdelim = '{';
		cRdelim = '}';
		break;

	case SINGLEQUOTE:
		cLdelim = '`';
		cRdelim = '\'';
		break;

	default:
	case UNKNOWNDELIM:
	case DEFAULTDELIM:
	case DOUBLEQUOTE:
		cLdelim = '"';
		cRdelim = '"';
		break;
	}
}

bool
HighParser::IsStringWithDelims(enum Delims Del)
{
	char cLdelim, cRdelim;
	SetDelims(Del, cLdelim, cRdelim);

	char cIn;
	if (skip_remarks(*this, *pIn, cIn)) {
		return false;
	}

	/* put back the first non-remark char */
	pIn->putback(cIn);

	/* if the left delimiter is found, true */
	return (cIn == cLdelim);
}

const char*
HighParser::GetStringWithDelims(enum Delims Del, bool escape)
{
	const char sFuncName[] = "HighParser::GetStringWithDelims()";

	if (CurrToken != HighParser::ARG) {
		silent_cerr("Parser error in "
			<< sFuncName << ", string arg expected at line "
			<< GetLineData() << std::endl);
		throw HighParser::ErrStringExpected(MBDYN_EXCEPT_ARGS);
	}

	char* s = sStringBuf;
	char* sTmp = s;

	char cLdelim, cRdelim;
	SetDelims(Del, cLdelim, cRdelim);

	char cIn;
	if (skip_remarks(*this, *pIn, cIn)) {
		return NULL;
	}

	/* Se trova il delimitatore sinistro, legge la stringa */
	if (cIn == cLdelim) {
		for (cIn = pIn->get(); cIn != cRdelim; cIn = pIn->get()) {
			/* Attenzione! cosi' la legge tutta,
			 * ma ne tiene solo iBufSize-1 caratteri */
			if (pIn->eof()) {
				/* FIXME: this should be an error ... */
				sTmp[0] = '\0';
				return s;

			} else if (sTmp < s + iDefaultBufSize - 1) {
				if (cIn == ESCAPE_CHAR) {
					cIn = pIn->get();
					if (cIn == '\n') {

						/*
						 * eat the newline as well, so that

							"first line\
							second line"

						 * actually results in "first linesecond line"
						 */

						cIn = pIn->get();

					} else if (cIn == '\r') {
						cIn = pIn->get();
						if (cIn != '\n') {
							pIn->putback(cIn);
							goto escaped_generic;
						}
						cIn = pIn->get();

					} else if (cIn == ESCAPE_CHAR) {
						if (!escape) {
							sTmp[0] = ESCAPE_CHAR;
							++sTmp;
						}

					} else {
escaped_generic:;
						if (escape) {
							int i, c = 0;
							char hex[3];

							/*
							 * allow non-printable chars in the form "\<hexpair>",
							 * so that "\78" is equivalent to "x";
							 * "\<non-hexpair>" is treated as an error.
							 */

							hex[0] = cIn;
							hex[1] = pIn->get();
							hex[2] = '\0';

							for (i = 0; i < 2; i++) {
								int shift = 4*(1 - i), h = 0;

								/* NOTE: this conversion relies
								 * on 0-9, a-f, A-F being consecutive,
								 * which is true for ASCII, but might
								 * not be for other encodings;
								 * bah, not critical right now */
								if (hex[i] >= '0' && hex[i]  <= '9') {
									h = hex[i] - '0';
								} else if (hex[i] >= 'a' && hex[i] <= 'f') {
									h = hex[i] - 'a';
								} else if (hex[i] >= 'A' && hex[i] <= 'F') {
									h = hex[i] - 'A';
								} else {
									silent_cerr("invalid escape sequence "
										"\"\\" << hex << "\" "
										"at line " << GetLineData()
										<< std::endl);
									throw ErrGeneric(MBDYN_EXCEPT_ARGS);
								}

								c += (h << shift);
							}
							cIn = c;

						} else {
							sTmp[0] = ESCAPE_CHAR;
							++sTmp;
						}
					}
				}
				sTmp[0] = cIn;
				++sTmp;
			}
		}

		/* Se trova una virgola o un punto e virgola, le rimette nello stream
		 * e passa oltre, restituendo un puntatore nullo. Il chiamante deve
		 * occuparsi della gestione del valore di default */
	} else if (cIn == ',' || cIn == ';') {
		pIn->putback(cIn);
		goto nullstring;

		/* Altrimenti c'e' qualcosa senza delimitatore. Adesso da' errore,
		 * forse e' piu' corretto fargli ritornare lo stream intatto */
	} else {
		silent_cerr("Parser error in "
			<< sFuncName << std::endl
			<< "first non-blank char at line "
			<< GetLineData() << " isn't a valid left-delimiter"
			<< std::endl);
		throw HighParser::ErrIllegalDelimiter(MBDYN_EXCEPT_ARGS);
	}

	/* Mette zero al termine della stringa */
	*sTmp = '\0';

nullstring:;
	NextToken(sFuncName);
	return s;
}


/* Legge un Vec3 */
Vec3
HighParser::GetVec3(void)
{
	Vec3 v(0.);
	return GetVec3(v);
}


/* Legge un Vec3 */
Vec3
HighParser::GetVec3(const Vec3& vDef)
{
	if (IsKeyWord("default")) {
		return vDef;
	}

	if (IsKeyWord("null")) {
		return Zero3;
	}

	doublereal x1 = GetReal(vDef(1));
	doublereal x2 = GetReal(vDef(2));
	doublereal x3 = GetReal(vDef(3));

	Vec3 v(x1, x2, x3);

	if (IsKeyWord("scale")) {
		v *= GetReal(1.);
	}

	return v;
}


/* Legge una matrice di orientazione sotto forma di due vettori (oppure eye) */
Mat3x3
HighParser::GetMatR2vec(void)
{
	if (IsKeyWord("eye")) {
		return Eye3;
	}

	if (IsKeyWord("matr")) {
		doublereal r11 = GetReal();
		doublereal r12 = GetReal();
		doublereal r13 = GetReal();
		doublereal r21 = GetReal();
		doublereal r22 = GetReal();
		doublereal r23 = GetReal();
		doublereal r31 = GetReal();
		doublereal r32 = GetReal();
		doublereal r33 = GetReal();

		/* FIXME: check for orthogonality? */
		return Mat3x3(r11, r21, r31, r12, r22, r32, r13, r23, r33);
	}

	if (IsKeyWord("euler" "parameters")) {
#if 0 /* FIXME: this function is TODO */
		doublereal e0 = GetReal();
		doublereal e1 = GetReal();
		doublereal e2 = GetReal();
		doublereal e3 = GetReal();

		return EulerParams2MatR(Vec3(e1, e2, e3));
#else
		silent_cerr("Line " << GetLineData()
			<< ": Euler parameters not implemented yet"
			<< std::endl);
		throw ErrGeneric(MBDYN_EXCEPT_ARGS);
#endif
	}

	if (IsKeyWord("euler") || IsKeyWord("euler" "123")) {
		doublereal e1 = GetReal();
		doublereal e2 = GetReal();
		doublereal e3 = GetReal();

		return EulerAngles123_2MatR(Vec3(e1, e2, e3));
	}

	if (IsKeyWord("euler" "313")) {
		doublereal e1 = GetReal();
		doublereal e2 = GetReal();
		doublereal e3 = GetReal();

		return EulerAngles313_2MatR(Vec3(e1, e2, e3));
	}

	if (IsKeyWord("euler" "321")) {
		doublereal e1 = GetReal();
		doublereal e2 = GetReal();
		doublereal e3 = GetReal();

		return EulerAngles321_2MatR(Vec3(e1, e2, e3));
	}

	if (IsKeyWord("vector")) {
		doublereal phi1 = GetReal();
		doublereal phi2 = GetReal();
		doublereal phi3 = GetReal();

		return RotManip::Rot(Vec3(phi1, phi2, phi3));
	}

	int i1 = GetInt();
	Vec3 v1;
	v1(1) = GetReal();
	v1(2) = GetReal();
	v1(3) = GetReal();

	int i2 = GetInt();
	Vec3 v2(0.);

	if (IsKeyWord("guess")) {
		int i_max = 1;
		int i_min = 1;

		if (fabs(v1(2)) > fabs(v1(1))) {
			i_max = 2;
		} else {
			i_min = 2;
		}

		if (fabs(v1(3)) > fabs(v1(i_max))) {
			i_max = 3;
		} else if (fabs(v1(3)) < fabs(v1(i_min))) {
			i_min = 3;
		}

		v2(i_min) = 1.;
		v2(i_max) = -v1(i_min)/v1(i_max);

	} else {
		v2(1) = GetReal();
		v2(2) = GetReal();
		v2(3) = GetReal();
	}

	return MatR2vec(i1, v1, i2, v2);
}


/* Legge una matrice 3x3 simmetrica come diagonale o triangolare superiore */
Mat3x3
HighParser::GetMat3x3Sym(void)
{
	if (IsKeyWord("null")) {
		return Mat3x3(0.);
	}
   
	Mat3x3 m;

	if (IsKeyWord("eye")) {
		m = Eye3;

	} else if (IsKeyWord("diag")) {
		doublereal m11 = GetReal();
		doublereal m22 = GetReal();
		doublereal m33 = GetReal();
		m = Mat3x3(m11, 0., 0., 0., m22, 0., 0., 0., m33);

	} else {   
		doublereal m11 = GetReal();
		doublereal m12 = GetReal();
		doublereal m13 = GetReal();
		doublereal m22 = GetReal();
		doublereal m23 = GetReal();
		doublereal m33 = GetReal();
		m = Mat3x3(m11, m12, m13, m12, m22, m23, m13, m23, m33);
	}

	if (IsKeyWord("scale")) {
		m *= GetReal(1.);
	}

	return m;
}

/* Legge una matrice 3x3 generica (diagonale o nulla) */
Mat3x3
HighParser::GetMat3x3(void)
{
	Mat3x3 m(0.);
	return GetMat3x3(m);
}


/* Legge una matrice 3x3 generica (diagonale o nulla) */
Mat3x3
HighParser::GetMat3x3(const Mat3x3& mDef)
{
	if (IsKeyWord("default")) {
		return mDef;

	} else if (IsKeyWord("null")) {
		return Zero3x3;
	}

	Mat3x3 m;

	if (IsKeyWord("eye")) {
		m = Eye3;

	} else if (IsKeyWord("diag")) {
		doublereal m11 = GetReal(mDef(1, 1));
		doublereal m22 = GetReal(mDef(2, 2));
		doublereal m33 = GetReal(mDef(3, 3));
		m = Mat3x3(m11, 0., 0., 0., m22, 0., 0., 0., m33);

	} else if (IsKeyWord("sym")) {
		doublereal m11 = GetReal(mDef(1, 1));
		doublereal m12 = GetReal(mDef(1, 2));
		doublereal m13 = GetReal(mDef(1, 3));
		doublereal m22 = GetReal(mDef(2, 2));
		doublereal m23 = GetReal(mDef(2, 3));
		doublereal m33 = GetReal(mDef(3, 3));
		m = Mat3x3(m11, m12, m13, m12, m22, m23, m13, m23, m33);

	} else if (IsKeyWord("skew")) {
		doublereal v1 = GetReal(mDef(3, 2));
		doublereal v2 = GetReal(mDef(1, 3));
		doublereal v3 = GetReal(mDef(2, 1));
		m = Mat3x3(0., -v3, v2, v3, 0., -v1, -v2, v1, 0.);

	} else {
		if (IsKeyWord("matr")) {
			/* eat it; not required */
			NO_OP;
		}
   
		doublereal m11 = GetReal(mDef(1, 1));
		doublereal m12 = GetReal(mDef(1, 2));
		doublereal m13 = GetReal(mDef(1, 3));
		doublereal m21 = GetReal(mDef(2, 1));
		doublereal m22 = GetReal(mDef(2, 2));
		doublereal m23 = GetReal(mDef(2, 3));
		doublereal m31 = GetReal(mDef(3, 1));
		doublereal m32 = GetReal(mDef(3, 2));
		doublereal m33 = GetReal(mDef(3, 3));
		m = Mat3x3(m11, m21, m31, m12, m22, m32, m13, m23, m33);
	}

	if (IsKeyWord("scale")) {
		m *= GetReal(1.);
	}

	return m;
}


/* Legge un Vec6 */
Vec6
HighParser::GetVec6(void)
{
	Vec6 v(0.);
	return GetVec6(v);
}


/* Legge un Vec6 */
Vec6
HighParser::GetVec6(const Vec6& vDef)
{
	if (IsKeyWord("null")) {
		return Zero6;
	}

	if (IsKeyWord("default")) {
		return vDef;
	}

	doublereal x1 = GetReal(vDef(1));
	doublereal x2 = GetReal(vDef(2));
	doublereal x3 = GetReal(vDef(3));
	doublereal x4 = GetReal(vDef(4));
	doublereal x5 = GetReal(vDef(5));
	doublereal x6 = GetReal(vDef(6));
	Vec6 v(x1, x2, x3, x4, x5, x6);

	if (IsKeyWord("scale")) {
		v *= GetReal(1.);
	}

	return v;
}


/* Legge una matrice 6x6 generica (diagonale o nulla) */
Mat6x6
HighParser::GetMat6x6(void)
{
	Mat6x6 m(0.);
	return GetMat6x6(m);
}


/* Legge una matrice 6x6 generica (diagonale o nulla) */
Mat6x6
HighParser::GetMat6x6(const Mat6x6& mDef)
{
	if (IsKeyWord("null")) {
		return Zero6x6;

	} else if (IsKeyWord("default")) {
		return mDef;
	}
   
	Mat6x6 m;

	if (IsKeyWord("eye")) {
		m = Eye6;

	} else if (IsKeyWord("diag")) {
		doublereal m11 = GetReal(mDef(1, 1));
		doublereal m22 = GetReal(mDef(2, 2));
		doublereal m33 = GetReal(mDef(3, 3));
		doublereal m44 = GetReal(mDef(4, 4));
		doublereal m55 = GetReal(mDef(5, 5));
		doublereal m66 = GetReal(mDef(6, 6));
		m = Mat6x6(
			m11, 0., 0., 0., 0., 0.,
			0., m22, 0., 0., 0., 0.,
			0., 0., m33, 0., 0., 0.,
			0., 0., 0., m44, 0., 0.,
			0., 0., 0., 0., m55, 0.,
			0., 0., 0., 0., 0., m66);
 
	} else if (IsKeyWord("sym")) {
		doublereal m11 = GetReal(mDef(1, 1));
		doublereal m12 = GetReal(mDef(1, 2));
		doublereal m13 = GetReal(mDef(1, 3));
		doublereal m14 = GetReal(mDef(1, 4));
		doublereal m15 = GetReal(mDef(1, 5));
		doublereal m16 = GetReal(mDef(1, 6));

		doublereal m22 = GetReal(mDef(2, 2));
		doublereal m23 = GetReal(mDef(2, 3));
		doublereal m24 = GetReal(mDef(2, 4));
		doublereal m25 = GetReal(mDef(2, 5));
		doublereal m26 = GetReal(mDef(2, 6));

		doublereal m33 = GetReal(mDef(3, 3));
		doublereal m34 = GetReal(mDef(3, 4));
		doublereal m35 = GetReal(mDef(3, 5));
		doublereal m36 = GetReal(mDef(3, 6));

		doublereal m44 = GetReal(mDef(4, 4));
		doublereal m45 = GetReal(mDef(4, 5));
		doublereal m46 = GetReal(mDef(4, 6));

		doublereal m55 = GetReal(mDef(5, 5));
		doublereal m56 = GetReal(mDef(5, 6));

		doublereal m66 = GetReal(mDef(6, 6));

		m = Mat6x6(
			m11, m12, m13, m14, m15, m16,
			m12, m22, m23, m24, m25, m26,
			m13, m23, m33, m34, m35, m36,
			m14, m24, m34, m44, m45, m46,
			m15, m25, m35, m45, m55, m56,
			m16, m26, m36, m46, m56, m66);

	} else if (IsKeyWord("anba")) {
		/* Formato ANBA, in cui vale la trasformazione:
		 * ex = e2
		 * ey = e3
		 * ez = e1
		 */
		doublereal m22 = GetReal(mDef(2, 2));
		doublereal m23 = GetReal(mDef(2, 3));
		doublereal m21 = GetReal(mDef(2, 1));
		doublereal m25 = GetReal(mDef(2, 5));
		doublereal m26 = GetReal(mDef(2, 6));
		doublereal m24 = GetReal(mDef(2, 4));

		doublereal m32 = GetReal(mDef(3, 2));
		doublereal m33 = GetReal(mDef(3, 3));
		doublereal m31 = GetReal(mDef(3, 1));
		doublereal m35 = GetReal(mDef(3, 5));
		doublereal m36 = GetReal(mDef(3, 6));
		doublereal m34 = GetReal(mDef(3, 4));

		doublereal m12 = GetReal(mDef(1, 2));
		doublereal m13 = GetReal(mDef(1, 3));
		doublereal m11 = GetReal(mDef(1, 1));
		doublereal m15 = GetReal(mDef(1, 5));
		doublereal m16 = GetReal(mDef(1, 6));
		doublereal m14 = GetReal(mDef(1, 4));

		doublereal m52 = GetReal(mDef(5, 2));
		doublereal m53 = GetReal(mDef(5, 3));
		doublereal m51 = GetReal(mDef(5, 1));
		doublereal m55 = GetReal(mDef(5, 5));
		doublereal m56 = GetReal(mDef(5, 6));
		doublereal m54 = GetReal(mDef(5, 4));

		doublereal m62 = GetReal(mDef(6, 2));
		doublereal m63 = GetReal(mDef(6, 3));
		doublereal m61 = GetReal(mDef(6, 1));
		doublereal m65 = GetReal(mDef(6, 5));
		doublereal m66 = GetReal(mDef(6, 6));
		doublereal m64 = GetReal(mDef(6, 4));

		doublereal m42 = GetReal(mDef(4, 2));
		doublereal m43 = GetReal(mDef(4, 3));
		doublereal m41 = GetReal(mDef(4, 1));
		doublereal m45 = GetReal(mDef(4, 5));
		doublereal m46 = GetReal(mDef(4, 6));
		doublereal m44 = GetReal(mDef(4, 4));

		m = Mat6x6(
			m11, m21, m31, m41, m51, m61,
			m12, m22, m32, m42, m52, m62,
			m13, m23, m33, m43, m53, m63,
			m14, m24, m34, m44, m54, m64,
			m15, m25, m35, m45, m55, m65,
			m16, m26, m36, m46, m56, m66);

	} else {
		if (IsKeyWord("matr")) {
			/* eat it; not required */
			NO_OP;
		}

		doublereal m11 = GetReal(mDef(1, 1));
		doublereal m12 = GetReal(mDef(1, 2));
		doublereal m13 = GetReal(mDef(1, 3));
		doublereal m14 = GetReal(mDef(1, 4));
		doublereal m15 = GetReal(mDef(1, 5));
		doublereal m16 = GetReal(mDef(1, 6));

		doublereal m21 = GetReal(mDef(2, 1));
		doublereal m22 = GetReal(mDef(2, 2));
		doublereal m23 = GetReal(mDef(2, 3));
		doublereal m24 = GetReal(mDef(2, 4));
		doublereal m25 = GetReal(mDef(2, 5));
		doublereal m26 = GetReal(mDef(2, 6));

		doublereal m31 = GetReal(mDef(3, 1));
		doublereal m32 = GetReal(mDef(3, 2));
		doublereal m33 = GetReal(mDef(3, 3));
		doublereal m34 = GetReal(mDef(3, 4));
		doublereal m35 = GetReal(mDef(3, 5));
		doublereal m36 = GetReal(mDef(3, 6));

		doublereal m41 = GetReal(mDef(4, 1));
		doublereal m42 = GetReal(mDef(4, 2));
		doublereal m43 = GetReal(mDef(4, 3));
		doublereal m44 = GetReal(mDef(4, 4));
		doublereal m45 = GetReal(mDef(4, 5));
		doublereal m46 = GetReal(mDef(4, 6));

		doublereal m51 = GetReal(mDef(5, 1));
		doublereal m52 = GetReal(mDef(5, 2));
		doublereal m53 = GetReal(mDef(5, 3));
		doublereal m54 = GetReal(mDef(5, 4));
		doublereal m55 = GetReal(mDef(5, 5));
		doublereal m56 = GetReal(mDef(5, 6));

		doublereal m61 = GetReal(mDef(6, 1));
		doublereal m62 = GetReal(mDef(6, 2));
		doublereal m63 = GetReal(mDef(6, 3));
		doublereal m64 = GetReal(mDef(6, 4));
		doublereal m65 = GetReal(mDef(6, 5));
		doublereal m66 = GetReal(mDef(6, 6));

		m = Mat6x6(
			m11, m21, m31, m41, m51, m61,
			m12, m22, m32, m42, m52, m62,
			m13, m23, m33, m43, m53, m63,
			m14, m24, m34, m44, m54, m64,
			m15, m25, m35, m45, m55, m65,
			m16, m26, m36, m46, m56, m66);
	}

	if (IsKeyWord("scale")) {
		m *= GetReal(1.);
	}

	return m;
}

/* provvisoria */
void
HighParser::GetMat6xN(Mat3xN& m1, Mat3xN& m2, integer iNumCols)
{
	ASSERT(iNumCols > 0);
	ASSERT(m1.iGetNumCols() == iNumCols);
	ASSERT(m2.iGetNumCols() == iNumCols);

	if (IsKeyWord("null")) {
		m1.Reset();
		m2.Reset();

	} else {
		int vi[] = { 1, 2, 3 };

		if (IsKeyWord("anba")) {
			vi[0] = 2;
			vi[1] = 3;
			vi[2] = 1;

		} else if (IsKeyWord("matr")) {
			/* eat it; not required */
			NO_OP;
		}

		for (int i = 0; i < 3; i++) {
			for (integer j = 1; j <= iNumCols; j++) {
				m1.Put(vi[i], j, GetReal());
			}
		}

		for (int i = 0; i < 3; i++) {
			for (integer j = 1; j <= iNumCols; j++) {
				m2.Put(vi[i], j, GetReal());
			}
		}

		if (IsKeyWord("scale")) {
			doublereal d = GetReal(1.);

			m1 *= d;
			m2 *= d;
		}
	}
}

doublereal
HighParser::Get(const doublereal& d)
{
	return GetReal(doublereal(d));
}

Vec3
HighParser::Get(const Vec3& v)
{
	return GetVec3(v);
}

Mat3x3
HighParser::Get(const Mat3x3& m)
{
	return GetMat3x3(m);
}

Vec6
HighParser::Get(const Vec6& v)
{
	return GetVec6(v);
}

Mat6x6
HighParser::Get(const Mat6x6& m)
{
	return GetMat6x6(m);
}

/* HighParser - end */

std::ostream&
operator << (std::ostream& out, const HighParser::ErrOut& err)
{
	out << err.iLineNumber;
		if (err.sFileName != NULL) {
			out << ", file <" << err.sFileName << '>';
		}
	return out;
}
